// This file belongs to each browser window.  It manages the timing issues 
// associated with checking for updates.  Only one browser window has control at a time.
// When a browser loads, it looks at all other open browsers (zero or more) and sees if 
// they are monitoring, if not, this browser starts monitors.
// When a controlling browser closes, it searches for the next open browser and passes on control.
// The only known bug associated with this control style is that during a transfer, all
// timers are reset.

// Thanks to RIP's Jerome O'Flaherty, whose code I used to get started

var pucConfig;
var pucIWatchPages = false;

var strings = null;


function puc_init() {

  // Localized strings
  strings = document.getElementById("puc-strings");

  window.removeEventListener("load", puc_init, true);

  // First, see if anyone else already is monitoring pages:
  var wm = Components.classes["@mozilla.org/appshell/window-mediator;1"].getService(Components.interfaces.nsIWindowMediator);
  var enumerator = wm.getEnumerator("navigator:browser");

  while(enumerator.hasMoreElements()) {
    var win = enumerator.getNext();
    if(win.pucIWatchPages) {
      // win has maintained the update checking, so we can copy any notifications already outstanding
      var statusBar = win.document.getElementById("status-bar");
      
      var menuItem = null;
      for(var i=0; (menuItem = statusBar.childNodes[i]); i++) {
        if(menuItem.page && menuItem.getAttribute("id") == "puc-notify-menu-" + menuItem.page.id) {
          addPageToNotifyIcon(menuItem.page);

          if(menuItem.hidden)
            pucHideNotifyIcon(menuItem.page.id);
        }
      }
      return; // We don't need to do any controlling or anything, just leave
    }
  }

  pucIWatchPages = true;

  // This opens PUC's configurations from the disk. It also makes sure preferences are up-to-date.
  pucConfig = new pucConfiguration();

  pucConfig.Initialize();
  puc_startAllTimers(true);
}


// This will be called by the controlling window if they close to a single
// remaining window so that we may take over control
// The only gross part about this is that the timers are restarted
function pucTakeOverWatch(newConfig)
{
  // It seems that destroy is called multiple times, and not always when we want it to be
  if(!pucIWatchPages) {
    //puc_logMessage("Taking over control");
    pucIWatchPages = true;
    pucCheckPreferences();

    // Most importantly, we hold the master copy of the configuration
    pucConfig = new pucConfiguration();
    pucConfig.copyInitialize(newConfig);
    puc_startAllTimers(false);
  }
}

// When we close a window, we must pass along control to another window
function puc_destroy()
{

  // Can't do this, it is unloading for some reason while it is loading!
  // window.removeEventListener("unload", puc_destroy, true); 

  if(pucIWatchPages) {
    var wm = Components.classes["@mozilla.org/appshell/window-mediator;1"].getService(Components.interfaces.nsIWindowMediator);
    var enumerator = wm.getEnumerator("navigator:browser");

    // If any BROSWER WINDOWS still exists
    if(enumerator.hasMoreElements()) {

      var win = enumerator.getNext();

      // Make sure we're not seeing ourself
      if(win == window)
        if(enumerator.hasMoreElements()) {
          win = enumerator.getNext();

        } else {

          // We were just seeing ourselves and no one is left, save the configuration to disk
          pucConfig.saveConfiguration();

          // and let Firefox clean up 
          return; 
        }

      win.pucTakeOverWatch(pucConfig);
    }
  }
}


// CheckNow is true when we first open Firefox, to accomodate the "checkAtStartup" option.
function puc_startAllTimers(checkNow)
{

  for(var i = 0; i < pucConfig.pages.length; i++) {
    if(pucConfig.pages[i].enabled) {
      puc_startTimer(pucConfig.pages[i]);
      if(pucConfig.pages[i].checkAtStartup && checkNow)
        pucConfig.pages[i].testForUpdate(puc_onNoPageChange, puc_onPageChange);
    }
  }
}


// One bug found (possibly only in win2k) was that when two or more interval timers are executed simultaneously, 
// they are both, for whatever reason, delayed for the length of one interval, at which time
// both calls execute.

// Note from Firefox source code:

    // The OS timer fired early (yikes!), and possibly out of order
    // too. Set |deadline| to be the time when the OS timer *should*
    // have fired so that any timers that *should* have fired before
    // aTimeout *will* be fired now. This happens most of the time on
    // Win2k.

// To circumvent this issue, we stagger when each timer is started based on the id of the page

function puc_startTimer(page) {
  setTimeout("puc_delayedStartTimer(pucConfig.getPageFromID(" + page.id + "));", (page.id%61)*100); // 61 is just some prime number
}

function puc_delayedStartTimer(page) {
  // Don't start a timer twice, so kill old running timers:
  if(page.timer != null)
    puc_stopTimer(page.timer);

  if(page.enabled) {
    var unitsToMilliseconds;
    switch(page.units) {
      case("seconds"):
        unitsToMilliseconds = 1000;
        break;
      case("minutes"):
        unitsToMilliseconds = 1000 * 60;
        break;
      case("hours"):
        unitsToMilliseconds = 1000 * 60 * 60;
        break;
      default:
        unitsToMilliseconds = 1000 * 60;
        puc_logMessage("Unit type " + page.units + " not recognized, defaulted to minutes.");
        break;
    }
    var interval = Math.round(page.period * unitsToMilliseconds);

    // We can assume that by this time the page has had its original source saved
    page.timer = setInterval("pucConfig.getPageFromID(" + page.id + ").testForUpdate(puc_onNoPageChange, puc_onPageChange)", interval);
    puc_logMessage("Started timer # " + page.timer + " for " + page.title + ": " + page.url + " (ID: " + page.id + ") at intervals of " + interval + " milliseconds");
  } 
}

function puc_stopTimer(timer) {
    
  clearInterval(timer);
  puc_logMessage("Timer # " + timer + " halted");
}


// Used by Options to reset the configuration
// Conveniently this also will load developer code
// within the copyInitialize function
function setPucConfig(newConfig)
{
  pucConfig = new pucConfiguration();
  pucConfig.copyInitialize(newConfig);
}


// This allows for the asynchronous http grabbing command checkForUpdate
// to be called from options just as it is closing.
function puc_checkForUpdateDelayed(page) {
  setTimeout("pucConfig.getPageFromID(" + page.id + ").testForUpdate(puc_onNoPageChange, puc_onPageChange)", 200);
}

function puc_onNoPageChange(page) {
  puc_logMessage("No change found in " + page.url);
  if(PrefUtils.getPref("puc.notifydespiteupdate"))
    alert(strings.getString("page.update.checker") + ":\n" + strings.getString("no.change.in") + " " + page.url +
    "\n(" + strings.getString("to.remove.message") + "\n" + strings.getString("prefs.directions") + ")");
}

function puc_onPageChange(page) {
  puc_logMessage(strings.getString("change.found.in") + " " + page.url);

  if(PrefUtils.getPref("puc.playsound")) {

    // Play the sound -- This code mostly taken from the popupsound extension

    // If I were less lazy, I'd cache the audio file and all that, but I don't use this feature, so screw it :-)
    
    var fileName = PrefUtils.getPref("puc.soundlocation");

    var player = Components.classes["@mozilla.org/sound;1"].getService(Components.interfaces.nsISound);
    var ios = Components.classes['@mozilla.org/network/io-service;1'].getService(Components.interfaces.nsIIOService);
  
    try {
      var file = Components.classes["@mozilla.org/file/local;1"].createInstance(Components.interfaces.nsILocalFile);
      file.initWithPath(fileName);
      player.play(ios.newFileURI(file));
    } catch(e) {
      try {
        player.play(ios.newURI(fileName, null, null));
      } catch(e2) {
        puc_logMessage("Sound Notification error:\n"+e2);
      }
    }
  }

  switch(PrefUtils.getPref("puc.notifymethod")) {
    case "dialog":
      // If the user accidentally has the page update every 5 seconds and they leave
      // this running, they won't come back to find a million boxes open because
      // "pucNotify"+page.id assures that strictly one window comes up for a given page,
  
      // First, see if this notification is already being displayed
      var isOpen = false;

      // Scan through Firefox windows, see if we're currently telling the user about the change
      var wm = Components.classes["@mozilla.org/appshell/window-mediator;1"].getService(Components.interfaces.nsIWindowMediator);
      var enumerator = wm.getEnumerator(""); // "" tells to look at ALL window types
      while(enumerator.hasMoreElements()) {
        var win = enumerator.getNext();
        if(win.name == "pucNotify"+page.id)
          isOpen = true;
      }
    
      if( !isOpen )
        window.openDialog("chrome://puc/content/pucNotify.xul", "pucNotify"+page.id, 
          "centerscreen,resizable", page, false); 
      break;

    case "icon":
      // Tell all browser windows to change their icon
      var wm = Components.classes["@mozilla.org/appshell/window-mediator;1"].getService(Components.interfaces.nsIWindowMediator);
      var enumerator = wm.getEnumerator("navigator:browser");

      while(enumerator.hasMoreElements()) {
        var win = enumerator.getNext();
        win.addPageToNotifyIcon(page);
      }

      break;
    default:
      break;
  }
}


function addPageToNotifyIcon(page) {

  var icon = window.document.getElementById("puc-notify-menu-"+page.id);
  
  if(icon) {
    pucShowNotifyIcon(page.id);
  } else {
    
    var defaultPageMenu = window.document.getElementById("puc-default-icon");
    var statusBar = window.document.getElementById("status-bar");
    var newPageMenu = defaultPageMenu.cloneNode(true);

    newPageMenu.page = page;
    newPageMenu.setAttribute("id", "puc-notify-menu-"+page.id);
    newPageMenu.setAttribute("label", page.title);
    newPageMenu.setAttribute("tooltiptext", "Update Found in " + page.title);
    newPageMenu.firstChild.firstChild.setAttribute("label", strings.getString("change.found.in") + " " + page.title);

    statusBar.appendChild(newPageMenu);
    pucShowNotifyIcon(page.id);
    return newPageMenu;
  }
}
